from __future__ import annotations
from datetime import datetime

import re

from scrummy.config import conf
from scrummy.constants import Constants
from scrummy.utils import leading_spaces, parse_line


class Epic:
    filepath: str
    frontmatter: dict
    todos: list[Todo]
    todos_dict: dict
    miscellanea: str

    def __init__(self):
        self.filepath = ""
        self.frontmatter = {}
        self.todos = []
        self.todos_dict = {}
        self.miscellanea = ""

    def __getitem__(self, key: str) -> Todo:
        return self.todos_dict[key]

    def __setitem__(self, key, value):
        self.todos_dict[key] = value

    def __str__(self) -> str:
        return self.render()

    def __repr__(self) -> str:
        return f"<Epic {self.filepath}>"

    def parse_frontmatter(self, frontmatter: list[str]) -> None:
        for line in frontmatter:
            key, value = parse_line(line, sep=":")
            self.frontmatter[key] = value

    def render(self, render_frontmatter: bool = True, render_miscellanea: bool = True) -> str:
        lines: list[str] = []
        if self.frontmatter and render_frontmatter:
            lines.append(f"---")
            for key, value in self.frontmatter.items():
                lines.append(f"{key}: {value}")
            lines.append("---")
            lines.append('\n')
        for todo in self.todos:
            lines.append(str(todo))
        lines.append(Constants.line_ending)
        if render_miscellanea:
            lines.append(self.miscellanea)
        return Constants.line_ending.join(lines)

    def update(self, todo: Todo, date: datetime) -> None:
        todo.date = date
        self[todo.id].completed = todo.completed
        self[todo.id].date = date

    def last_of_depth(self, depth: int) -> Todo | None:
        if depth <= 0:
            if self.todos:
                return self.todos[-1]
            else:
                return None
        todo = self.todos[-1]
        while depth > todo.depth:
            if todo.children:
                todo = todo.children[-1]
            else:
                break
        return todo


class Todo:
    id: str = ""
    epic_id: int | None = None
    content: str = ""
    date: datetime | None = None
    completed: bool = False
    parent: Todo = None
    depth: int = 0
    children: list[Todo]

    def __init__(self, lines: list[str]):
        """
        Parses a multi-line todo of the form:

        ```
        * [ ] 234-4 <content>
        ```
        Parameters
        ----------
        lines: List[str]
        """
        self.children = []
        if lines[0].startswith("  "):
            pass
        self.depth = leading_spaces(lines[0]) // conf.indent_size
        self.completed = bool(re.search(r'\[x]', lines[0]))
        ident = re.search(r'\d+-\d+', lines[0])
        if ident:
            self.id = ident.group()
            self.epic_id = int(self.id.split('-')[0])
            end_of_id = ident.span()[-1]
        else:
            end_of_id = Constants.opener_length
        content: str = ''.join([lines[0][end_of_id:], *lines[1:]]).replace('\n', '')
        self.content = re.sub(r' +', ' ', content)

    def __str__(self):
        return self.render()

    def __repr__(self):
        return f"<Todo {self.id}>"

    def count_parents(self) -> int:
        """
        Counts the number of parents (i.e., the count of indentations to prepend)

        Returns
        -------
        int
        """
        if self.parent:
            return self.parent.count_parents() + 1
        return 0

    def render(self):
        """
        Renders a single todo to a string.

        Produces a string of the form:

        `- [ ] 234-4 <content>` if completed is False
        `- [x] 234-4 <content>` if completed is True

        The dash is replaced by configured list_indicator

        The indentation is determined by the depth of the todo (i.e., how many parents it has).

        The string is broken into multiple lines if it is too long (configured by indent_size).

        Returns
        -------
        str
        """
        completed: str = 'x' if self.completed else ' '
        ident: str = self.id if self.id else ''
        date: str = f' ({self.date.strftime(conf.date_format)})' if self.date else ''

        # '- [ ] 234-4 <content>'
        if self.completed:
            single_line: str = f'{conf.list_indicator} [{completed}] {ident}{self.content}{date}'
        else:
            single_line: str = f'{conf.list_indicator} [{completed}] {ident}{self.content}'

        indentation: str = ' ' * conf.indent_size * self.count_parents()
        split_strings: list[str] = []
        idx: int = 0
        first_line: bool = True
        while single_line[idx:].strip():
            end = idx + conf.max_line_length - len(indentation)
            if len(single_line[idx:end]) == end - idx:
                end = single_line[idx:end].rfind(' ')
            if first_line:
                split_strings.append(indentation + single_line[idx:end].strip())
                first_line = False
            else:
                split_strings.append(indentation + ' ' * Constants.opener_length + single_line[idx:end].strip())
            idx = end
        for child in self.children:
            split_strings.append(child.render())
        return Constants.line_ending.join(split_strings)



from typing import Any
from pathlib import Path


def leading_spaces(line: str) -> int:
    """
    Calculates the number of leading spaces in a string.

    Parameters
    ----------
    line

    Returns
    -------
    int
    """
    return len(line) - len(line.lstrip())


def parse_value(value: str) -> Any:
    """
    Parses a value into its proper type.

    Parameters
    ----------
    value: str

    Returns
    -------
    Any
    """
    value = value.strip()
    if value.isdigit():
        return int(value)
    try:
        return float(value)
    except ValueError:
        pass
    if value.lower() in ['true', 'false']:
        return value.lower() == 'true'
    if value.lower() in ['none', 'null']:
        return None
    if value.startswith('"') and value.endswith('"'):
        value = value[1:-1]
    if value.startswith("'") and value.endswith("'"):
        value = value[1:-1]
    if '~' in value:
        value = str(Path(value).expanduser())
    return value


def parse_line(line: str, sep: str = '=') -> tuple[str, Any] | None:
    """
    Parses a line into a key and value.

    Examples
    --------
    `key = value`
    `key: value`

    Parameters
    ----------
    line: str
    sep: str
        The separator for splitting

    Returns
    -------
    Tuple[str, Any]
        The key and value of the line.
    """
    if line.startswith('#'):
        return None
    if sep not in line:
        return None
    key, value = line.split(sep, 1)
    value: Any = parse_value(value)
    return key, value



import os
from pathlib import Path
from typing import Any
from dataclasses import dataclass

from xdg import xdg_config_home

from scrummy.utils import parse_line

config_root: str = os.path.join(xdg_config_home(), 'scrummy')
config_file: str = os.path.join(config_root, 'scrummyrc')


@dataclass
class Config:
    """
    Config object for the scrummy application.
    """
    home: Path = Path('~/documents/scrummy').expanduser()
    todo_filename: str = 'todo.md'
    max_line_length: int = 80
    indent_size: int = 2
    date_format: str = '%Y/%m/%d'
    list_indicator: str = '-'

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)

    def __setitem__(self, key: Any, value: Any):
        if hasattr(self, key):
            setattr(self, key, value)
        else:
            raise KeyError(f'{key} is not a valid configuration key.')

    @property
    def todo_file(self) -> str:
        return os.path.join(self.home, self.todo_filename)


def init_config() -> Config:
    """
    Initializes the config object.

    Returns
    -------
    Config
        The config object.
    """
    if os.path.exists(config_file):
        with open(config_file, 'rt') as f:
            data = f.readlines()
    else:
        return Config()
    config = Config()
    for line in data:
        try:
            key, val = parse_line(line)
            config[key] = val
        except TypeError:
            pass
    return config


# The main config object for the application.
conf: Config = init_config()



"""
1. Read in todo.md.
2. All numbered done tasks are marked done in scrum/.
3. All undone tasks are kept in the new todo.md.
4. All done tasks are deleted from todo.md.
5. Everything else is ignored.
"""
import os
import re
import shutil
from datetime import datetime, timedelta
from pathlib import Path

import dateutil as du
from dateutil.parser import ParserError

from scrummy.config import conf
from scrummy.constants import Constants
from scrummy.models import Todo, Epic


def link_todo(todo: Todo, epic: Epic):
    last_todo = epic.last_of_depth(todo.depth-1)
    if last_todo and last_todo.depth < todo.depth:
        last_todo.children.append(todo)
        todo.parent = last_todo
    else:
        epic.todos.append(todo)
    if todo.id:
        epic[todo.id] = todo


def parse_todofile(filename: Path) -> tuple[datetime | None, Epic]:
    """
    Parse a todo file into a list of Todo objects.

    Parameters
    ----------
    filename: str | Path

    Returns
    -------
    tuple[datetime|None, Epic]
    """
    with open(filename, 'rt') as f:
        items = f.readlines()
    current_line: list[str] = []
    is_todo: bool = False
    is_frontmatter: bool = False
    file_date: datetime | None = None
    frontmatter: list[str] = []
    miscellanea: list[str] = []
    have_date: bool = False

    epic = Epic()
    epic.filepath = filename

    for line in items:
        # getting frontmatter
        if line.strip() == '---':
            is_frontmatter = not is_frontmatter
        elif is_frontmatter:
            frontmatter.append(line)
        # trying to get the date, assuming we're above the todo list.
        elif line.startswith('#') and not have_date:
            try:
                file_date = du.parser.parse(line[1:].strip())
                have_date = True
            except ParserError:
                pass
        elif re.match(Constants.opener_regex, line.lstrip()):
            have_date = True  # We assume that the first todo is after the date (in case it doesn't exist)
            if current_line:
                link_todo(Todo(current_line), epic)
                current_line = []
            current_line.append(line)
            is_todo = True
            if line.strip() == items[-1].strip():
                link_todo(Todo(current_line), epic)
                current_line = []
        elif is_todo:
            if not line.strip():
                link_todo(Todo(current_line), epic)
                current_line = []
                is_todo = False
            else:
                current_line.append(line)
        elif have_date:
            miscellanea.append(line)
    epic.parse_frontmatter(frontmatter)
    epic.miscellanea = ''.join(miscellanea)
    epic.filepath = filename
    return file_date, epic


def update_epics(epics: dict[str, Epic], sprint: Epic, date: datetime | None):
    keep = []
    for todo in sprint.todos:
        if todo.epic_id in epics:
            epics[todo.epic_id].update(todo, date)
        if not todo.completed:
            keep.append(todo)
    sprint.todos = keep
    sprint.todos_dict = {todo.id: todo for todo in sprint.todos}


def rollover_todo():
    date, sprint = parse_todofile(Path(conf.todo_file))
    epics: dict[str, Epic] = {}
    for path in Path(os.path.join(conf.home, 'scrum')).glob('*.md'):
        _, epic = parse_todofile(path)
        epics[epic.frontmatter['id']] = epic
    update_epics(epics, sprint, date)
    for epic in epics.values():
        with open(epic.filepath, 'wt') as f:
            f.write(str(epic))
    shutil.move(sprint.filepath, os.path.join(conf.home, 'sprints', date.strftime('%Y-%m-%d') + '.md'))
    with open(Path(conf.todo_file), 'wt') as f:
        # write the date as a top headline
        f.write('# ' + (datetime.now() + timedelta(days=1)).strftime('%Y-%m-%d') + '\n\n')
        f.write(sprint.render(render_frontmatter=False, render_miscellanea=False))


if __name__ == '__main__':
    rollover_todo()



import pathlib
from setuptools import setup, find_packages

from scrummy import __version__

HERE: pathlib.Path = pathlib.Path(__file__).parent
README: str = (HERE / "README.md").read_text()

setup(
    name='scrummy',
    version=__version__.__version__,
    description=__version__.__description__,
    long_description=README,
    long_description_content_type="text/markdown",
    author=__version__.__author__,
    author_email=__version__.__author_email__,
    license=__version__.__license__,
    url=__version__.__url__,
    project_urls={
        'Homepage': __version__.__url__,
        'Source': __version__.__url__,
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=['click', 'python-dateutil', 'xdg'],
    entry_points={
        'console_scripts': [
            'scrummy = scrummy.main:cli',
        ]
    }
)
